
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <helianthus.h>


#define FRIENDS_COUNT       3
#define ENEMIES_COUNT       2

#define PLAYER_SPEED        200
#define PLAYER_ROT_SPEED    200
#define PLAYER_ARROW_SPEED  400
#define PLAYER_ARROW_DELAY  0.4
#define PLAYER_HITS         15
#define PLAYER_WIN_BONUS    5

#define FRIEND_SPEED        200
#define FRIEND_ROT_SPEED    200
#define FRIEND_ARROW_SPEED  400
#define FRIEND_ARROW_DELAY  0.4
#define FRIEND_HITS         10

#define ENEMY_SPEED         150
#define ENEMY_ROT_SPEED     150
#define ENEMY_ARROW_SPEED   300
#define ENEMY_ARROW_DELAY   0.5
#define ENEMY_HITS          10

#define MENU_FRIENDS_COUNT  1
#define MENU_ENEMIES_COUNT  1


struct Ship;
typedef void (*ShipFunc)(struct Ship *ship);
typedef struct Ship {
  Sprite sprite;

  double startX;
  double speed;
  double rotSpeed;
  double arrowSpeed;
  double arrowDelay;
  Animation arrowAnim;

  struct Ship *enemies;
  int enemiesCount;

  int hits;
  struct Ship *target;
  double tx;
  double ty;

  double maxMove;
  double maxRot;
  double arrowTime;

  ShipFunc onUpdate;
  ShipFunc onKill;
} Ship;


Animation Enemyon;
Animation Friendon;
Animation Playeron;

Animation EnemyArrowon;
Animation FriendArrowon;
Animation PlayerArrowon;

Animation MenuEnemyon;
Animation MenuFriendon;
Animation MenuEnemyArrowon;
Animation MenuFriendArrowon;

Animation gameOveron;
Animation gameOverbg;
Animation menuon;
Animation menuContinueon;
Animation menuRadioactiveon;
Animation menuFearon;
Animation menuQuiton;

Animation bg;


Sound gameSounds[2];
Sound gameOverSound;

Group arrows;
int gameMode;
int wins;
int menu;
int selectedGameSound;

Ship friends[FRIENDS_COUNT];
Ship enemies[ENEMIES_COUNT];

Group menuArrows;
Group menuShips;
Ship menuFriends[MENU_FRIENDS_COUNT];
Ship menuEnemies[MENU_ENEMIES_COUNT];


void gameStop();

void shipUpdateBot(Ship *s);
void shipUpdateMenuBot(Ship *s);
void shipKillEnemy(Ship *s);
void shipKillFriend(Ship *s);
void shipUpdatePlayer(Ship *s);
void shipKillPlayer(Ship *s);


double calcDr(double dx, double dy, double r) {
  double dr = atan2(dy, dx)*180/PI - r;
  while(dr >  180) dr -= 360;
  while(dr < -180) dr += 360;
  return dr;
}


void shipInitEnemy(Ship *s, double startX) {
  memset(s, 0, sizeof(*s));
  s->sprite = createSpriteEx(startX, 12.5, 25, 25, Enemyon);
  spriteSetRotation(s->sprite, 90);
  spriteSetDepth(s->sprite, 1);

  s->startX = startX;
  s->speed = ENEMY_SPEED;
  s->rotSpeed = ENEMY_ROT_SPEED;
  s->arrowSpeed = ENEMY_ARROW_SPEED;
  s->arrowDelay = ENEMY_ARROW_DELAY;
  s->arrowAnim = EnemyArrowon;

  s->hits = ENEMY_HITS;

  s->enemies = friends;
  s->enemiesCount = FRIENDS_COUNT;
  if (s->enemiesCount)
    s->target = &s->enemies[randomNumber(0, s->enemiesCount-1)];

  s->onUpdate = shipUpdateBot;
  s->onKill = shipKillEnemy;
}

void shipInitFriend(Ship *s, double startX) {
  shipInitEnemy(s, startX);
  spriteSetAnimation(s->sprite, Friendon);
  spriteSetY(s->sprite, 512 - 12.5);
  spriteSetRotation(s->sprite, -90);

  s->speed = FRIEND_SPEED;
  s->rotSpeed = FRIEND_ROT_SPEED;
  s->arrowSpeed = FRIEND_ARROW_SPEED;
  s->arrowDelay = FRIEND_ARROW_DELAY;
  s->arrowAnim = FriendArrowon;

  s->hits = FRIEND_HITS;

  s->enemies = enemies;
  s->enemiesCount = ENEMIES_COUNT;
  if (s->enemiesCount)
    s->target = &s->enemies[randomNumber(0, s->enemiesCount-1)];

  s->onKill = shipKillFriend;
}

void shipInitPlayer(Ship *s, double startX) {
  shipInitFriend(s, startX);
  spriteSetAnimation(s->sprite, Playeron);

  s->speed = PLAYER_SPEED;
  s->rotSpeed = PLAYER_ROT_SPEED;
  s->arrowSpeed = PLAYER_ARROW_SPEED;
  s->arrowDelay = PLAYER_ARROW_DELAY;
  s->arrowAnim = PlayerArrowon;

  s->hits = PLAYER_HITS;

  s->onUpdate = shipUpdatePlayer;
  s->onKill = shipKillPlayer;
}

void shipInitMenuEnemy(Ship *s, double x, double y) {
  memset(s, 0, sizeof(*s));
  s->sprite = createSpriteEx(x, y, 25, 25, MenuEnemyon);
  spriteSetRotation(s->sprite, 90);
  spriteSetDepth(s->sprite, 1);
  groupAdd(menuShips, s->sprite);

  s->arrowSpeed = ENEMY_ARROW_SPEED;
  s->arrowDelay = ENEMY_ARROW_DELAY;
  s->arrowAnim = MenuEnemyArrowon;

  s->hits = ENEMY_HITS;

  s->enemies = menuFriends;
  s->enemiesCount = MENU_FRIENDS_COUNT;
  s->onUpdate = shipUpdateMenuBot;
}

void shipInitMenuFriend(Ship *s, double x, double y) {
  memset(s, 0, sizeof(*s));
  s->sprite = createSpriteEx(x, y, 25, 25, MenuFriendon);
  spriteSetRotation(s->sprite, -90);
  spriteSetDepth(s->sprite, 1);
  groupAdd(menuShips, s->sprite);

  s->arrowSpeed = FRIEND_ARROW_SPEED;
  s->arrowDelay = FRIEND_ARROW_DELAY;
  s->arrowAnim = MenuFriendArrowon;

  s->hits = FRIEND_HITS;

  s->enemies = menuEnemies;
  s->enemiesCount = MENU_ENEMIES_COUNT;
  s->onUpdate = shipUpdateMenuBot;
}

void shipDeinit(Ship *s) {
  spriteDestroy(s->sprite);
  memset(s, 0, sizeof(*s));
}

void shipUpdate(Ship *s, double dt) {
  s->maxMove = s->speed*dt;
  s->maxRot = s->rotSpeed*dt;
  s->arrowTime -= dt;
  if (s->onUpdate && s->hits > 0)
    s->onUpdate(s);
}

void shipHit(Ship *s) {
  if (s->hits > 0) {
    --s->hits;
    if (s->hits <= 0 && s->onKill)
      s->onKill(s);
  }
}

void shipMove(Ship *s, double m) {
  if (m >  s->maxMove) m =  s->maxMove;
  if (m < -s->maxMove) m = -s->maxMove;
  s->maxMove -= fabs(m);
  spriteMoveToDirection(s->sprite, m, spriteGetRotation(s->sprite));
  double x = spriteGetX(s->sprite);
  double y = spriteGetY(s->sprite);
  if (x <     12.5) x = 12.5;
  if (x > 512-12.5) x = 512-12.5;
  if (y <     12.5) y = 12.5;
  if (y > 512-12.5) y = 512-12.5;
  spriteSetXY(s->sprite, x, y);
}

void shipRotate(Ship *s, double r) {
  if (r >  s->maxRot) r =  s->maxRot;
  if (r < -s->maxRot) r = -s->maxRot;
  s->maxRot -= fabs(r);
  spriteSetRotation(s->sprite, spriteGetRotation(s->sprite) + r);
}

void shipShoot(Ship *s) {
  if (s->arrowTime <= 0) {
    double x = spriteGetX(s->sprite);
    double y = spriteGetY(s->sprite);
    double r = spriteGetRotation(s->sprite);
    Sprite arrow = createSpriteEx(x, y, 25, 25, s->arrowAnim);
    spriteSetRotation(arrow, r);
    spriteSetSpeedAndDirection(arrow, s->arrowSpeed, r);
    spriteSetUserData(arrow, s);
    groupAdd(arrows, arrow);
    spriteDestroyTimer(arrow, 5);
    s->arrowTime = s->arrowDelay;
  }
}


void shipUpdateBot(Ship *s) {
  double x = spriteGetX(s->sprite);
  double y = spriteGetY(s->sprite);
  double r = spriteGetRotation(s->sprite);
  if (s->target) {
    s->tx = spriteGetX(s->target->sprite);
    s->ty = spriteGetY(s->target->sprite);
  }
  double dx = s->tx - x;
  double dy = s->ty - y;
  double dr = calcDr(dx, dy, r);
  double dist = sqrt(dx*dx + dy*dy);

  shipRotate(s, dr);
  shipMove(s, dist);

  x = spriteGetX(s->sprite);
  y = spriteGetY(s->sprite);
  r = spriteGetRotation(s->sprite);
  for(int i = 0; i < s->enemiesCount; ++i) {
    double ex = spriteGetX(s->enemies[i].sprite);
    double ey = spriteGetY(s->enemies[i].sprite);
    if (fabs(calcDr(ex - x, ey - y, r)) < 5) shipShoot(s);
  }

  if (s->target) {
    if (dist < 50) {
      s->tx = randomFloat()*(512-25) + 12.5;
      s->ty = randomFloat()*(512-25) + 12.5;
      s->target = NULL;
    }
  } else {
    if (dist < 20 && s->enemiesCount > 0) {
      s->target = &s->enemies[ randomNumber(0, s->enemiesCount-1) ];
    }
  }
}

void shipUpdateMenuBot(Ship *s) {
  spritePointTo(s->sprite, mouseX(), mouseY());
  double x = spriteGetX(s->sprite);
  double y = spriteGetY(s->sprite);
  double r = spriteGetRotation(s->sprite);
  for(int i = 0; i < s->enemiesCount; ++i) {
    double ex = spriteGetX(s->enemies[i].sprite);
    double ey = spriteGetY(s->enemies[i].sprite);
    if (fabs(calcDr(ex - x, ey - y, r)) < 5) {
      if (s->arrowTime <= 0) {
        Sprite arrow = createSpriteEx(x, y, 25, 25, s->arrowAnim);
        spriteSetRotation(arrow, r);
        spriteSetSpeedAndDirection(arrow, s->arrowSpeed, r);
        spriteSetUserData(arrow, s);
        groupAdd(menuArrows, arrow);
        spriteDestroyTimer(arrow, 5);
        s->arrowTime = s->arrowDelay;
      }
    }
  }
}

void shipUpdatePlayer(Ship *s) {
  if (keyDown( "left"  )) shipRotate(s, -1000);
  if (keyDown( "right" )) shipRotate(s,  1000);
  if (keyDown( "up"    )) shipMove(s,  1000);
  if (keyDown( "down"  )) shipMove(s, -1000);
  if (keyDown( "space" )) shipShoot(s);
}


void shipKillEnemy(Ship *s) {
  spriteSetXY(s->sprite, s->startX, 12.5);
  spriteSetRotation(s->sprite, 90);
  if (s->enemiesCount > 0)
    s->target = &s->enemies[ randomNumber(0, s->enemiesCount-1) ];
  s->hits = ENEMY_HITS;
  friends[0].hits += PLAYER_WIN_BONUS;
  if (friends[0].hits > PLAYER_HITS) friends[0].hits = PLAYER_HITS;
  ++wins;
}

void shipKillFriend(Ship *s) {
  spriteSetXY(s->sprite, s->startX, 512-12.5);
  spriteSetRotation(s->sprite, -90);
  if (s->enemiesCount > 0)
    s->target = &s->enemies[ randomNumber(0, s->enemiesCount-1) ];
  s->hits = FRIEND_HITS;
  --wins;
}

void shipKillPlayer(Ship *s) {
  gameStop();
}



void gameStop() {
  if (gameMode == 2) return;
  soundStop(gameSounds[selectedGameSound]);
  if (gameMode == 1) {
    for(int i = 0; i < ENEMIES_COUNT; ++i)
      shipDeinit(&enemies[i]);
    for(int i = 0; i < FRIENDS_COUNT; ++i)
      shipDeinit(&friends[i]);
  }
  groupDestroyEach(arrows);
  soundPlay(gameOverSound, TRUE);
  gameMode = 2;
}

void gameStart() {
  if (gameMode == 1) gameStop();

  wins = 0;
  soundStop(gameOverSound);
  for(int i = 0; i < ENEMIES_COUNT; ++i)
    shipInitEnemy(&enemies[i], 512.0/(ENEMIES_COUNT+1)*(i+1));
  int p = FRIENDS_COUNT/2;
  for(int i = 0; i < FRIENDS_COUNT; ++i) {
    double x = 512.0/(FRIENDS_COUNT+1)*(i+1);
    if (i == p) shipInitPlayer(&friends[0], x);
      else shipInitFriend(&friends[i < p ? i+1 : i], x);
  }
  soundPlay(gameSounds[selectedGameSound], TRUE);

  gameMode = 1;
}

void setMenu(int m) {
  if (!m == !menu) return;
  menu = !!m;
  groupSetFrozenEach(arrows, menu);
}



void init() {
  background(COLOR_BLACK);
  
  bg = createAnimation("data/background.png");
  
  Enemyon = createAnimation("data/Enemy.png");
  Friendon = createAnimation("data/Friend.png");
  Playeron = createAnimation("data/Player.png");

  EnemyArrowon = createAnimation("data/arrow.png");
  FriendArrowon = createAnimation("data/FriendArrow.png");
  PlayerArrowon = createAnimation("data/FriendArrow.png");

  MenuEnemyon = createAnimation("data/Enemy c..png");
  MenuFriendon = createAnimation("data/Friend c..png");
  MenuEnemyArrowon = createAnimation("data/arrow.png");
  MenuFriendArrowon = createAnimation("data/FriendArrow.png");

  gameOveron = createAnimation("data/game-over.png");
  gameOverbg = createAnimation("data/game-over-bg.png");

  menuon            = createAnimation("data/Menu.png");
  menuContinueon    = createAnimation("data/Continue.png");
  menuRadioactiveon = createAnimation("data/Radioactive.png");
  menuFearon        = createAnimation("data/Fear.png");
  menuQuiton        = createAnimation("data/Quit.png");

  gameSounds[0] = createSound("data/03_fear.ogg");
  gameSounds[1] = createSound("data/Radioactive - Imagine Dragons.ogg");
  gameOverSound = createSound("data/Mission to Mars.ogg");

  arrows = createGroup();
  
  menuArrows = createGroup();
  menuShips = createGroup();
  shipInitMenuEnemy(&menuEnemies[0], 256-64, 256-64);
  shipInitMenuFriend(&menuFriends[0], 256+64, 256-64);
  
  gameStart();
}

void draw() {
  saveState();
  double dt = windowGetFrameTime();

  rectTextured(bg,0,0,512,512);  
  
  if (gameMode == 1) {
    if (menu) {
      for(int i = 0; i < MENU_ENEMIES_COUNT; ++i)
        shipUpdate(&menuEnemies[i], dt);
      for(int i = 0; i < MENU_FRIENDS_COUNT; ++i)
        shipUpdate(&menuFriends[i], dt);

      for(int i = groupGetCount(menuArrows) - 1; i >= 0; --i) {
        Sprite a = groupGet(menuArrows, i);
        Ship *ship = (Ship*)spriteGetUserData(a);
        for(int j = groupGetCount(menuShips) - 1; j >= 0; --j) {
          Sprite s = groupGet(menuShips, j);
          if (ship->sprite != s && spriteOverlap(a, s))
            { spriteDestroy(a); break; }
        }
      }
    } else {
      for(int i = 0; i < ENEMIES_COUNT; ++i)
        shipUpdate(&enemies[i], dt);
      for(int i = 0; i < FRIENDS_COUNT; ++i)
        shipUpdate(&friends[i], dt);

      for(int i = groupGetCount(arrows) - 1; i >= 0; --i) {
        Sprite s = groupGet(arrows, i);
        void *d = spriteGetUserData(s);
        for(int j = 0; j < FRIENDS_COUNT; ++j) {
          Ship *ship = &friends[j];
          if (d != ship && spriteOverlap(s, ship->sprite)) {
            spriteDestroy(s);
            d = s = NULL;
            shipHit(ship);
            break;
          }
        }
        if (gameMode != 1) break;
        if (!s) continue;
        for(int j = 0; j < ENEMIES_COUNT; ++j) {
          Ship *ship = &enemies[j];
          if (d != ship && spriteOverlap(s, ship->sprite)) {
            spriteDestroy(s);
            d = s = NULL;
            shipHit(ship);
            break;
          }
        }
        if (gameMode != 1) break;
      }
    }

    groupSetVisibleEach(menuArrows, FALSE);
    groupSetVisibleEach(menuShips, FALSE);
    drawSprites();
  }

  if (gameMode == 1) {
    noFill();
    stroke(COLOR_WHITE);
    textAlign(HALIGN_LEFT, VALIGN_BOTTOM);
    textf(5, 512 - 5, "%d", friends[0].hits);
    textAlign(HALIGN_RIGHT, VALIGN_BOTTOM);
    textf(512 - 5, 512 - 5, "%d", wins);
    textAlign(HALIGN_LEFT, VALIGN_BOTTOM);
    textSize(10);
    for(int i = 0; i < ENEMIES_COUNT; ++i) {
      double x = spriteGetX(enemies[i].sprite);
      double y = spriteGetY(enemies[i].sprite);
      textf(x + 5, y - 5, "%d", enemies[i].hits);
    }
    for(int i = 1; i < FRIENDS_COUNT; ++i) {
      double x = spriteGetX(friends[i].sprite);
      double y = spriteGetY(friends[i].sprite);
      textf(x + 5, y - 5, "%d", friends[i].hits);
    }
  } else
  if (gameMode == 2) {
    noStroke();
    fill(COLOR_BLACK);
    rect(0, 0, 512, 512);
    fill(colorByRGBA( 1, 1, 1, (sin(windowGetSeconds()) + 1)/2 ));
    rectTextured(gameOverbg, 0, 0, 512, 512);
    rectTextured(gameOveron, 0, 0, 512, 512);

    noFill();
    stroke(COLOR_WHITE);
    textAlign(HALIGN_RIGHT, VALIGN_BOTTOM);
    textf(512 - 5, 512 - 5, "%d", wins);
    if (keyDown("any return")) gameStart();
  }

  if (menu) {
    noStroke();
    fill(colorByRGBA(0, 0, 0, 0.75));
    rect(0, 0, 512, 512);
    
    groupSetVisibleEach(menuArrows, TRUE);
    groupSetVisibleEach(menuShips, TRUE);
    groupDraw(menuArrows);
    groupDraw(menuShips);
    
    fill(COLOR_WHITE);
    rectTextured(menuon,            206, 223, 100, 65);
    rectTextured(menuContinueon,    211, 228,  90, 15);
    rectTextured(selectedGameSound ? menuFearon : menuRadioactiveon,
                                    211, 248,  90, 15);
    rectTextured(menuQuiton,        211, 268,  90, 15);
    double mx = mouseX();
    double my = mouseY();
    if (mouseWentDown("left") && mx >= 211 && mx < 301) {
      if (my >= 228 && my < 228+15) setMenu(FALSE);
      if (my >= 248 && my < 248+15) {
        soundStop(gameSounds[selectedGameSound]);
        selectedGameSound = !selectedGameSound;
        if (gameMode == 1) soundPlay(gameSounds[selectedGameSound], TRUE);
      }
      if (my >= 268 && my < 268+15) windowStop();
    }
    if (keyWentDown("escape")) setMenu(FALSE);
  } else {
    if (keyWentDown("escape")) setMenu(TRUE);
  }

  restoreState();
}

int main() {
  windowSetTitle("Space Combat 2.1");
  windowSetVariableFrameRate();
  windowSetInit(&init);
  windowSetDraw(&draw);
  windowRun();
  return 0;
}
